#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from random import randint
import sys
import os
from glxeveloop.fps import FPS

# Require when you haven't GLXBob as default Package
current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.dirname(current_dir))


# Unittest
class TestTimer(unittest.TestCase):
    def setUp(self):
        # Before the test start
        self.fps = FPS()
        self.fps.debug = True

    def test_default_values(self):
        self.assertEqual(self.fps.value, 60.0)
        self.assertEqual(self.fps.max, float("inf"))
        self.assertEqual(self.fps.min, 30.0)
        self.assertEqual(self.fps.fps_increment, 1.0)
        self.assertEqual(self.fps.fps_min_increment, 0.1)
        self.assertEqual(self.fps.fps_max_increment, 100.0)

    def test_value(self):
        self.fps.value = 42
        self.assertEqual(float(42), self.fps.value)

        self.fps.value = 42.42
        self.assertEqual(42.42, self.fps.value)

        self.fps.min = 3
        self.fps.max = 4

        self.fps.value = 42.42
        self.assertEqual(4.0, self.fps.value)

        self.fps.value = 1
        self.assertEqual(3.0, self.fps.value)

        self.assertRaises(TypeError, setattr, self.fps, "value", "Hello.42")

    def test_min_fps(self):
        self.fps.min = 42
        self.assertEqual(float(42), self.fps.min)

        self.fps.min = 42.42
        self.assertEqual(42.42, self.fps.min)

        self.assertRaises(TypeError, setattr, self.fps, "min", "Hello.42")

    def test_max_fps(self):
        self.fps.max = 42
        self.assertEqual(float(42), self.fps.max)

        self.fps.max = 42.42
        self.assertEqual(42.42, self.fps.max)

        self.assertRaises(TypeError, setattr, self.fps, "max", "Hello.42")

    def test_fps_increment(self):
        random_value = randint(1, 250)
        self.fps.fps_increment = float(random_value)
        self.assertEqual(self.fps.fps_increment, float(random_value))
        random_value = randint(1, 250)
        self.assertRaises(
            TypeError, setattr, self.fps, "fps_increment", int(random_value)
        )

    def test_min_fps_increment(self):
        random_value = randint(1, 250)
        self.fps.fps_min_increment = float(random_value)
        self.assertEqual(self.fps.fps_min_increment, float(random_value))

        random_value = randint(1, 250)
        self.assertRaises(
            TypeError, setattr, self.fps, "fps_min_increment", int(random_value)
        )

    def test_max_fps_increment(self):
        random_value = float(randint(1, 250))
        self.fps.fps_max_increment = random_value
        self.assertEqual(self.fps.fps_max_increment, random_value)

        random_value = randint(1, 250)
        self.assertRaises(
            TypeError, setattr, self.fps, "fps_max_increment", int(random_value)
        )


if __name__ == "__main__":
    unittest.main()



import unittest

from glxeveloop.hooks import Hooks


def fake_function():
    pass


class TestMainLoopHooks(unittest.TestCase):
    def setUp(self) -> None:
        self.hooks = Hooks()
        self.hooks.debug = True

    def test_statement(self):
        self.assertIsNone(self.hooks.statement)
        self.hooks.statement = fake_function
        self.assertEqual(fake_function, self.hooks.statement)
        self.hooks.statement = None
        self.assertIsNone(self.hooks.statement)
        self.assertRaises(TypeError, setattr, self.hooks, "statement", "Hello.42")

    def test_parsing(self):
        self.assertIsNone(self.hooks.parsing)
        self.hooks.parsing = fake_function
        self.assertEqual(fake_function, self.hooks.parsing)
        self.hooks.parsing = None
        self.assertIsNone(self.hooks.parsing)
        self.assertRaises(TypeError, setattr, self.hooks, "parsing", "Hello.42")

    def test_pre(self):
        self.assertIsNone(self.hooks.pre)
        self.hooks.pre = fake_function
        self.assertEqual(fake_function, self.hooks.pre)
        self.hooks.pre = None
        self.assertIsNone(self.hooks.pre)
        self.assertRaises(TypeError, setattr, self.hooks, "pre", "Hello.42")

    def test_cmd(self):
        self.assertIsNone(self.hooks.cmd)
        self.hooks.cmd = fake_function
        self.assertEqual(fake_function, self.hooks.cmd)
        self.hooks.cmd = None
        self.assertIsNone(self.hooks.cmd)
        self.assertRaises(TypeError, setattr, self.hooks, "cmd", "Hello.42")

    def test_post(self):
        self.assertIsNone(self.hooks.post)
        self.hooks.post = fake_function
        self.assertEqual(fake_function, self.hooks.post)
        self.hooks.post = None
        self.assertIsNone(self.hooks.post)
        self.assertRaises(TypeError, setattr, self.hooks, "post", "Hello.42")

    def test_finalization(self):
        self.assertIsNone(self.hooks.finalization)
        self.hooks.finalization = fake_function
        self.assertEqual(fake_function, self.hooks.finalization)
        self.hooks.finalization = None
        self.assertIsNone(self.hooks.finalization)
        self.assertRaises(TypeError, setattr, self.hooks, "finalization", "Hello.42")

    def test_dispatch(self):
        self.assertIsNone(self.hooks.dispatch)
        self.hooks.dispatch = fake_function
        self.assertEqual(fake_function, self.hooks.dispatch)
        self.hooks.dispatch = None
        self.assertIsNone(self.hooks.dispatch)
        self.assertRaises(TypeError, setattr, self.hooks, "dispatch", "Hello.42")

    def test_keyboard_interruption(self):
        self.assertIsNone(self.hooks.keyboard_interruption)
        self.hooks.keyboard_interruption = fake_function
        self.assertEqual(fake_function, self.hooks.keyboard_interruption)
        self.hooks.keyboard_interruption = None
        self.assertIsNone(self.hooks.keyboard_interruption)
        self.assertRaises(
            TypeError, setattr, self.hooks, "keyboard_interruption", "Hello.42"
        )


if __name__ == "__main__":
    unittest.main()



from glxeveloop.timer import Timer


class TimerProperty:
    def __init__(self) -> None:
        self.__timer = Timer()

    @property
    def timer(self):
        return self.__timer

    @timer.setter
    def timer(self, value):
        if value is None:
            value = Timer()
        if not isinstance(value, Timer):
            raise TypeError("'timer' property value must be a Timer instance or None")
        if self.timer != value:
            self.__timer = value



class RunningProperty:
    debug: bool

    def __init__(self) -> None:
        self.__running = False

    @property
    def running(self):
        """
        The property :py:data:`running`, store the actual running state rate of the MainLoop, it property is use by
        the class :class:`MainLoop <glxeveloop.loop.mainloop.MainLoop>`.

        **Default Value:** ``false``

        :return: :py:data:`running` property.
        :rtype: :py:data:`bool`
        :raise TypeError: if parameter is not a :py:data:`bool` type or None
        """
        return self.__running

    @running.setter
    def running(self, value):
        """
        Set the is_running attribute
        """
        if value is None:
            value = False
        if not isinstance(value, bool):
            raise TypeError("'running' property value must be bool type or None")
        if self.running != value:
            self.__running = value



from glxeveloop.loop.singleton import Singleton
from glxeveloop.loop.builder import Builder


class MainLoop(metaclass=Singleton):
    @property
    def loop(self):
        return self.__loop

    def __init__(self):
        self.__loop = Builder().loop()

    # def handle_event(self):
    #     pass



import unittest

from glxeveloop.properties.debug import DebugProperty


class TestDebug(unittest.TestCase):
    def setUp(self) -> None:
        self.debug = DebugProperty()

    def test_debug(self):
        self.assertEqual(False, self.debug.debug)
        self.debug.debug = None
        self.assertEqual(False, self.debug.debug)

        self.debug.debug = True
        self.assertEqual(True, self.debug.debug)
        self.assertRaises(TypeError, setattr, self.debug, "debug", "Hello.42")


if __name__ == "__main__":
    unittest.main()



import unittest

from glxeveloop.loop.mainloop import MainLoop
from glxeveloop.hooks import Hooks
from glxeveloop.timer import Timer
from queue import Queue


class FakeApplication(object):
    def __init__(self):
        self.permit_keyboard_interruption = True
        self.count = 1

    @staticmethod
    def get_mouse():
        return 0, 0, 0, 0, 0

    def eveloop_precmd(self):
        if self.count >= 10:
            MainLoop().loop.stop()

    def eveloop_cmd(self):
        print(
            "COUNT:{0} FPS:{1} TIME:{2} sec".format(
                self.count,
                MainLoop().loop.timer.fps.value,
                MainLoop().loop.timer.fps.value / 60,
            )
        )

    def eveloop_postcmd(self):
        self.count += 1

    def eveloop_finalization(self):
        pass

    def events_dispatch(self, detailed_signal, args):
        pass

    def gen_a_exception(self):
        raise Exception

    @staticmethod
    def gen_a_statement():
        return 1

    def dispatch(self, detailed_signal, args):
        pass


class TestMainLoop(unittest.TestCase):
    def setUp(self) -> None:
        self.mainloop = MainLoop().loop
        # print(self.mainloop)
        # self.mainloop.debug = True

    def test_debug(self):
        self.mainloop.debug = False
        self.assertFalse(self.mainloop.debug)
        # self.assertFalse(self.mainloop.timer.debug)
        # self.assertFalse(self.mainloop.hooks.debug)

        self.mainloop.debug = True
        self.assertTrue(self.mainloop.debug)
        # self.assertTrue(self.mainloop.timer.debug)
        # self.assertTrue(self.mainloop.hooks.debug)

        # self.assertRaises(TypeError, setattr, self.mainloop, "debug", "Hello.42")

    def test_hooks(self):
        self.assertTrue(isinstance(self.mainloop.hooks, Hooks))

        old_hooks = self.mainloop.hooks

        self.mainloop.hooks = None
        self.assertTrue(isinstance(self.mainloop.hooks, Hooks))

        self.assertNotEqual(old_hooks, self.mainloop.hooks)

        self.assertRaises(TypeError, setattr, self.mainloop, "hooks", "Hello.42")

    def test_timer(self):
        self.assertTrue(isinstance(self.mainloop.timer, Timer))
        old_timer = self.mainloop.timer

        self.mainloop.timer = None
        self.assertTrue(isinstance(self.mainloop.timer, Timer))

        self.assertNotEqual(old_timer, self.mainloop.timer)

        self.assertRaises(TypeError, setattr, self.mainloop, "timer", "Hello.42")

    def test_queue(self):
        self.mainloop.application = FakeApplication()
        old_object = self.mainloop.queue
        self.mainloop.queue = Queue()
        self.assertTrue(isinstance(self.mainloop.queue, Queue))
        self.assertNotEqual(old_object, self.mainloop.queue)

        self.mainloop.queue = None
        self.assertTrue(isinstance(self.mainloop.queue, Queue))

        self.assertRaises(TypeError, setattr, self.mainloop, "queue", 42)

    def test_running(self):
        self.assertFalse(MainLoop().loop.running)

        self.mainloop.running = True
        self.assertTrue(self.mainloop.running)

        self.mainloop.running = None
        self.assertFalse(self.mainloop.running)

        self.assertRaises(TypeError, setattr, self.mainloop, "running", 42)

    def test_start(self):
        self.mainloop.debug = True

        app = FakeApplication()
        self.mainloop.hooks.statement = app.gen_a_statement
        self.mainloop.timer.fps.value = 5.0
        self.mainloop.timer.fps.max = 5.0
        self.mainloop.hooks.cmd = app.eveloop_cmd
        self.mainloop.hooks.pre = app.eveloop_precmd
        self.mainloop.hooks.post = app.eveloop_postcmd
        self.mainloop.hooks.finalization = app.eveloop_finalization
        self.mainloop.start()

        self.mainloop.hooks.cmd = app.gen_a_exception
        self.assertRaises(Exception, self.mainloop.start)

    def test_stop(self):
        self.mainloop.debug = True

        self.assertFalse(self.mainloop.running)
        self.mainloop.running = True
        self.assertTrue(self.mainloop.running)

        self.mainloop.stop()
        self.assertFalse(self.mainloop.running)

    def test_handle_event(self):
        self.mainloop.debug = True
        self.mainloop.queue.put(["Hello.42", (1, 2, 3)])
        app = FakeApplication()

        self.mainloop.hooks.dispatch = app.dispatch
        self.mainloop.handle_event()


if __name__ == "__main__":
    unittest.main()



import unittest

from io import StringIO
from unittest.mock import patch
from glxeveloop import Bus
from glxeveloop.loop.mainloop import MainLoop


def hello42(handler, signal_details, args):
    print(args)


class FakeApplication(Bus):
    def __init__(self):
        Bus.__init__(self)
        self.permit_keyboard_interruption = True

    @staticmethod
    def get_mouse():
        return 0, 0, 0, 0, 0

    def eveloop_precmd(self):
        pass

    def eveloop_cmd(self):
        pass

    def eveloop_postcmd(self):
        pass

    def eveloop_finalization(self):
        pass

    def eveloop_dispatch(self, detailed_signal, args):
        pass


class TestBus(unittest.TestCase):
    def setUp(self):
        self.bus = Bus()
        self.bus.debug = True
        self.mainloop = MainLoop()

    def test_subscriptions(self):

        self.assertEqual({}, self.bus.subscriptions)
        self.bus.subscriptions = {"Hello": 42}
        self.assertEqual({"Hello": 42}, self.bus.subscriptions)

        self.bus.subscriptions = None
        self.assertEqual({}, self.bus.subscriptions)
        self.assertRaises(TypeError, setattr, self.bus, "subscriptions", "Hello.42")

    def test_connect(self):

        self.bus.connect("HELLO42", hello42, "hello", "42")

        self.assertEqual(hello42, self.bus.subscriptions["HELLO42"][0])
        self.assertEqual(("hello", "42"), self.bus.subscriptions["HELLO42"][1])

    def test_disconnect(self):

        self.assertEqual(0, len(self.bus.subscriptions))
        self.bus.connect("HELLO42", hello42, "hello", "42")
        self.bus.connect("HELLO43", hello42, "hello", "43")

        self.assertEqual(hello42, self.bus.subscriptions["HELLO42"][0])
        self.assertEqual(hello42, self.bus.subscriptions["HELLO43"][0])
        self.assertEqual(("hello", "42"), self.bus.subscriptions["HELLO42"][1])
        self.assertEqual(("hello", "43"), self.bus.subscriptions["HELLO43"][1])
        self.assertEqual(2, len(self.bus.subscriptions))

        self.bus.disconnect("HELLO42", hello42)
        self.assertEqual(1, len(self.bus.subscriptions))

        self.assertEqual(hello42, self.bus.subscriptions["HELLO43"][0])
        self.assertEqual(("hello", "43"), self.bus.subscriptions["HELLO43"][1])
        self.bus.disconnect("HELLO43", hello42)
        self.assertEqual(0, len(self.bus.subscriptions))

    def test_emit(self):
        self.assertEqual(0, MainLoop().loop.queue.qsize())
        self.bus.emit("HELLO42", {"Hello": "42"})
        self.assertEqual(1, MainLoop().loop.queue.qsize())
        event = MainLoop().loop.queue.get()
        self.assertEqual("HELLO42", event[0])
        self.assertEqual({"Hello": "42"}, event[1])

    def test_event_flush(self):

        self.assertEqual(0, len(self.bus.subscriptions))
        self.bus.connect("HELLO42", hello42)
        with patch("sys.stdout", new=StringIO()) as fake_out:
            self.bus.events_flush("HELLO42", ("hello", "42"))
            expected = "('hello', '42')\n"
            self.assertEqual(fake_out.getvalue(), expected)

    def test_events_dispatch(self):
        fake_app = FakeApplication()
        self.assertEqual(0, len(fake_app.subscriptions))
        fake_app.connect("HELLO42", hello42)
        with patch("sys.stdout", new=StringIO()) as fake_out:
            fake_app.events_dispatch("HELLO42", ("hello", "42"))
            expected = "('hello', '42')\n"
            self.assertEqual(fake_out.getvalue(), expected)


if __name__ == "__main__":
    unittest.main()



from glxeveloop.hooks import Hooks
from glxeveloop.loop.handle_events import HandleEvents


class SequenceLoop:
    hooks: Hooks
    handle_event = HandleEvents().handle_event

    def sequence(self) -> None:
        """
        * Parse user input into a Statement object
        * Start timer
        * Call precmd method
        * Add statement to History
        * Call cmd method
        * Call postcmd method
        * Stop timer and display the elapsed time
        * In Case of Exit call methods loop_finalization
        """
        try:
            self.hooks.pre()
        except TypeError:  # pragma: no cover
            pass

        try:
            self.handle_event()
        except TypeError:  # pragma: no cover
            pass

        try:
            self.hooks.cmd()
        except TypeError:  # pragma: no cover
            pass

        try:
            self.hooks.post()
        except TypeError:  # pragma: no cover
            pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# It script it publish under GNU GENERAL PUBLIC LICENSE
# http://www.gnu.org/licenses/gpl-3.0.en.html
# Author: the Galaxie EveLoop Team, all rights reserved

import logging
from glxeveloop.properties.debug import DebugProperty


class Hooks(DebugProperty):
    def __init__(self):
        DebugProperty.__init__(self)

        self.__statement = None
        self.__parsing = None
        self.__pre = None
        self.__cmd = None
        self.__post = None
        self.__finalization = None
        self.__dispatch = None
        self.__keyboard_interruption = None

        self.statement = None
        self.parsing = None
        self.pre = None
        self.cmd = None
        self.post = None
        self.finalization = None
        self.dispatch = None
        self.keyboard_interruption = None

    @property
    def statement(self):
        """
        Register a hook to be called before parsing information, support to iterate the application

        :return: A callable function
        :rtype: callable
        """
        return self.__statement

    @statement.setter
    def statement(self, value):
        """
        Set the ``statement`` property value

        :param value: A callable function
        :type value: callable
        :raise TypeError: When ``statement`` property is not a ``callable`` or None
        """
        if value and not hasattr(value, "__call__"):
            raise TypeError("'statement' property value must be a callable")

        if value != self.statement:
            if self.debug:
                logging.debug("%s", f"{self.__class__.__name__}: [STATEMENT] {value}")
            self.__statement = value

    @property
    def parsing(self):
        """
        Register a hook to be called after have received ``statement`` from the application.

        :return: A callable function
        :rtype: callable
        """
        return self.__parsing

    @parsing.setter
    def parsing(self, value):
        """
        Set the ``parsing`` property value

        :param value: A callable function
        :type value: callable
        :raise TypeError: When ``parsing`` property is not a ``callable`` or None
        """
        if value and not hasattr(value, "__call__"):
            raise TypeError("'parsing' property value must be a callable")

        if value != self.parsing:
            if self.debug:
                logging.debug("%s", f"{self.__class__.__name__}: [PARSING] {value}")
            self.__parsing = value

    @property
    def pre(self):
        """
        Register a hook to be called before the command function.

        :return: A callable function
        :rtype: callable
        """
        return self.__pre

    @pre.setter
    def pre(self, value):
        """
        Set the ``cmd`` property value

        :param value: A callable function
        :type value: callable
        :raise TypeError: When ``pre`` property is not a ``callable`` or None
        """
        if value and not hasattr(value, "__call__"):
            raise TypeError("'pre' property value must be a callable")

        if value != self.pre:
            if self.debug:
                logging.debug("%s", f"{self.__class__.__name__}: [PRE] {value}")
            self.__pre = value

    @property
    def cmd(self):
        """
        Register a hook to be called as the command function.

        :return: A callable function
        :rtype: callable
        """
        return self.__cmd

    @cmd.setter
    def cmd(self, value):
        """
        Set the ``cmd`` property value

        :param value: A callable function
        :type value: callable
        :raise TypeError: When ``cmd`` property is not a ``callable`` or None
        """
        if value and not hasattr(value, "__call__"):
            raise TypeError("'cmd' property value must be a callable")

        if value != self.cmd:
            if self.debug:
                logging.debug("%s", f"{self.__class__.__name__}: [CMD] {value}")
            self.__cmd = value

    @property
    def post(self):
        """
        Register a hook to be called after the command function.

        :return: A callable function
        :rtype: callable
        """
        return self.__post

    @post.setter
    def post(self, value):
        """
        Set the ``post`` property value

        :param value: A callable function
        :type value: callable
        :raise TypeError: When ``post`` property is not a ``callable`` or None
        """
        if value and not hasattr(value, "__call__"):
            raise TypeError("'post' property value must be a callable")

        if value != self.post:
            if self.debug:
                logging.debug("%s", f"{self.__class__.__name__}: [POST] {value}")
            self.__post = value

    @property
    def finalization(self):
        """
        Register a hook to be called just before mainloop exit, whether it completes successfully or not.

        :return: A callable function
        :rtype: callable
        """
        return self.__finalization

    @finalization.setter
    def finalization(self, value):
        """
        Set the ``finalization`` property value

        :param value: A callable function
        :type value: callable
        :raise TypeError: When ``finalization`` property is not a ``callable`` or None
        """
        if value and not hasattr(value, "__call__"):
            raise TypeError("'finalization' property value must be a callable")

        if value != self.finalization:
            if self.debug:
                logging.debug(
                    "%s", f"{self.__class__.__name__}: [FINALIZATION] {value}"
                )
            self.__finalization = value

    @property
    def dispatch(self):
        """
        Should be remove soon

        :return: A callable function
        :rtype: callable
        """
        return self.__dispatch

    @dispatch.setter
    def dispatch(self, value):
        """
        Set the ``dispatch`` property value

        :param value: A callable function
        :type value: callable
        :raise TypeError: When ``dispatch`` property is not a ``callable`` or None
        """
        if value and not hasattr(value, "__call__"):
            raise TypeError("'dispatch' property value must be a callable")

        if value != self.dispatch:
            if self.debug:
                logging.debug("%s", f"{self.__class__.__name__}: [DISPATCH] {value}")
            self.__dispatch = value

    @property
    def keyboard_interruption(self):
        """
        Register a hooks it Shortcut KeyboardInterrupt during the mainloop execution

        :return: A callable function
        :rtype: callable
        """
        return self.__keyboard_interruption

    @keyboard_interruption.setter
    def keyboard_interruption(self, value):
        """
        Set the ``keyboard_interruption`` property value

        :param value: A callable function
        :type value: callable
        :raise TypeError: When ``keyboard_interruption`` property is not a ``callable`` or None
        """
        if value and not hasattr(value, "__call__"):
            raise TypeError("'keyboard_interruption' property value must be a callable")

        if value != self.keyboard_interruption:
            if self.debug:
                logging.debug(
                    "%s", f"{self.__class__.__name__}: [KEYBOARD_INTERRUPTION] {value}"
                )
            self.__keyboard_interruption = value



from glxeveloop.hooks import Hooks


class HooksProperty:
    debug: bool

    def __init__(self) -> None:
        self.__hooks = Hooks()

    @property
    def hooks(self):
        return self.__hooks

    @hooks.setter
    def hooks(self, value):
        if value is None:
            value = Hooks()
        if not isinstance(value, Hooks):
            raise TypeError("'hooks' property value must be a Hooks instance or None")
        if value != self.hooks:
            self.__hooks = value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from glxeveloop.loop.mainloop import MainLoop
from glxeveloop.bus import Bus

APPLICATION_VERSION = "0.1.5"
APPLICATION_AUTHORS = ["Tuuuux", "Mo"]

__all__ = [
    "Bus",
    "MainLoop",
]

mainloop = MainLoop()



import logging
import sys
from glxeveloop.hooks import Hooks
from glxeveloop.timer import Timer
from glxeveloop.loop.sequence import SequenceLoop
from glxeveloop.loop.stop import Stop


class Start:
    running: bool
    debug: bool
    timer: Timer
    hooks: Hooks
    sequence: SequenceLoop().sequence
    stop: Stop().stop

    def start(self) -> None:
        """
        Runs a Loop until ``Mainloop.stop()`` is called on the loop. If this is called for the thread of the loop's
        , it will process queue from the loop, otherwise it will simply wait.
        """
        if self.debug:
            logging.debug("%s", f"{self.__class__.__name__}: Starting...")
        self.running = True

        # Normally it the first refresh of the application, it can be considered as the first stdscr display.
        # Consider a chance to crash before the start of the loop
        try:
            try:
                self.sequence()
            except TypeError:  # pragma: no cover
                pass

        except Exception:
            try:
                self.stop()
            except TypeError:  # pragma: no cover
                pass
            sys.stdout.write("{0}\n".format(sys.exc_info()[0]))
            sys.stdout.flush()
            raise

        # A bit light for notify about we are up and running, but we are really inside the main while(1) loop
        try:
            if self.debug:
                logging.debug("%s", f"{self.__class__.__name__}: Started")
        except TypeError:  # pragma: no cover
            pass
        # The loop
        while self.running:

            try:
                try:
                    self.hooks.statement()
                except TypeError:  # pragma: no cover
                    pass

                try:
                    self.sequence()
                except TypeError:  # pragma: no cover
                    pass

                try:
                    self.timer.tick()
                except TypeError:  # pragma: no cover
                    pass

            except KeyboardInterrupt:  # pragma: no cover
                try:
                    self.hooks.keyboard_interruption()
                except TypeError:
                    if hasattr(self, "stop") and self.stop:
                        self.stop()

            except Exception:  # pragma: no cover
                try:
                    self.stop()
                except TypeError:
                    pass
                sys.stdout.write("{0}\n".format(sys.exc_info()[0]))
                sys.stdout.flush()
                raise

        try:
            if self.debug:
                logging.debug(
                    "%s", f"{self.__class__.__name__}: Call finalization method"
                )

            self.hooks.finalization()

            if self.debug:
                logging.debug(
                    "%s", f"{self.__class__.__name__}: All operations is stop"
                )

        except TypeError:  # pragma: no cover
            pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-


class DebugProperty:
    def __init__(self):
        self.__debug = False

    @property
    def debug(self):
        """
        Debugging level of information's.

        Generally it highly stress the console and is here for future maintenance of that Application.
        """
        return self.__debug

    @debug.setter
    def debug(self, value=None):
        """
        Debugging level of information's.

        Generally it highly stress the console and is here for future maintenance of that Application.

        :param value: True is debugging mode is enabled, False for disable it.
        :type value: bool
        :raise TypeError: when "debug" argument is not a :py:__area_data:`bool`
        """
        if value is None:
            value = False
        if not isinstance(value, bool):
            raise TypeError('"debug" must be a boolean type or None')
        if self.debug != value:
            self.__debug = value
            if hasattr(self, "fps"):
                if self.fps:
                    self.fps.debug = self.debug



from queue import Queue


class QueueProperty:
    def __init__(self) -> None:
        self.__queue = Queue()

    @property
    def queue(self):
        return self.__queue

    @queue.setter
    def queue(self, value):
        if value is None:
            value = Queue()
        if not isinstance(value, Queue):
            raise TypeError("'queue' property value must be a Queue instance or None")
        if value != self.queue:
            self.__queue = value



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
from time import time, sleep
from queue import Queue
from glxeveloop.fps import FPS
from glxeveloop.properties.debug import DebugProperty


# Reference Document: http://code.activestate.com/recipes/579053-high-precision-fps/
class Timer(DebugProperty):
    SPEED_STEP_PERCENTAGE = 10
    DEFAULT_BE_FAST_MULTIPLICATOR = 100
    MIN_BE_FAST_MULTIPLICATOR = 0

    """
    :Description:

    The :class:`Timer <glxeveloop.timer.Timer>` object contain a self-correcting timing algorithms.

    That self-correcting timing algorithms have associated property's it permit to control the
    :class:`Timer <glxeveloop.timer.Timer>` object.

    The power saving happen that because the loop try to limit the number of mainloop cycle
    The :class:`Timer <glxeveloop.timer.Timer>` object update value itself and can be requested, via internal method's.
    """

    def __init__(self):
        DebugProperty.__init__(self)
        # Internal

        self.__queue = None
        self.__fps = None
        self.__time_departure = None
        self.__be_fast = None
        self.__be_fast_multiplication = None

        # First Init

        self.fps = None
        self.be_fast = None
        self.be_fast_multiplication = None
        self.position = 0
        self.queue = Queue(maxsize=10)
        # self.queue.maxsize = self.size

    @property
    def fps(self):
        """
        The FPS object

        :return: :py:data:`fps` property value. (in **fps**)
        :rtype: FPS
        """
        return self.__fps

    @fps.setter
    def fps(self, value=None):
        """
        Set the :class:`Timer <glxeveloop.timer.Timer>` :py:data:`fps` property.

        :param value: a FPS object
        :type value: FPS
        :raise TypeError: if ``fps`` parameter is not a  FPS instance
        """
        if value is None:
            value = FPS()
        if not isinstance(value, FPS):
            raise TypeError("'fps' property value must be a FPS instance")
        if self.fps != value:
            self.__fps = value

    @property
    def queue(self):
        """
        Store the Timer Buffer object.

        :return: TimerBuffer instance
        :rtype: Memory
        """
        return self.__queue

    @queue.setter
    def queue(self, value=None):
        if value is None:
            value = Queue()
        if not isinstance(value, Queue):
            raise TypeError("'queue' property value must be a Memory instance or None")
        if self.queue != value:
            self.__queue = value

    def tick(self):
        """
        Return :py:obj:`True` or :py:obj:`False` "when necessary" , that mean according with all the self-correcting
        timing algorithms and they configuration property's


        .. code-block:: python

           timer = Timer()
           while True:
               # Do stuff that might take significant time here

               if timer.tick():
                   print('Hello World!')

        :return: :py:obj:`True` when it's time or :py:obj:`False` if a adjustment job of :py:data:`fps` property
                 should be done
        :rtype: bool
        """
        if self.time_departure is None:
            self.time_departure = self.time

        # Increase position
        self.position += 1

        # The algorithm
        target = self.position / self.fps.value
        passed = self.time - self.time_departure
        differ = target - passed

        # Reset time reference due to time variation
        # Should never be remove or for a true system if compensate time variation
        if self.queue.full():
            self.time_departure = self.time
            self.position = 1

            # Determine a increment factor for fast convergence
            # half_sum = sum(self.queue.buffer[:len(self.queue.buffer) / 2])
            half_sum = 0.0
            rest_sum = 0.0
            orig_size = self.queue.qsize()

            while not self.queue.empty():
                if self.queue.qsize() <= int(orig_size / 2):
                    half_sum += self.queue.get()
                else:
                    rest_sum += self.queue.get()

            # It's time to analyze the result
            # First Check if that equal
            if int(half_sum) == int(rest_sum):
                self.be_fast_multiplication = 0
                self.be_fast = False
                if self.debug:
                    logging.info(
                        "%s",
                        f"{self.__class__.__name__}:[GOAL]-> Increment {self.fps.fps_increment} fps, {self.fps.value} fps",
                    )

            # Check if we have to down fps
            elif half_sum < rest_sum:
                if self.be_fast:
                    if self.debug:
                        logging.info(
                            "%s",
                            f"{self.__class__.__name__}:[DOWN]-> Increment {self.fps_accelerated} fps, {self.fps.value} fps",
                        )
                    self.be_fast_multiplication -= self.SPEED_STEP_PERCENTAGE

                else:
                    self.be_fast_multiplication = None
                    if self.debug:
                        logging.info(
                            "%s",
                            f"{self.__class__.__name__}:[DOWN]-> Increment {self.fps.fps_increment} fps, {self.fps.value} fps",
                        )
                self.be_fast = False

            # Check if we can speed up
            elif half_sum > rest_sum:
                if self.be_fast:
                    if self.debug:
                        logging.info(
                            "%s",
                            f"{self.__class__.__name__}:[UP]-> Increment {self.fps_accelerated} fps, {self.fps.value} fps",
                        )
                    self.be_fast_multiplication += self.SPEED_STEP_PERCENTAGE

                else:
                    if self.debug:
                        logging.info(
                            "%s",
                            f"{self.__class__.__name__}:[UP]-> Increment {self.fps.fps_increment} fps, {self.fps.value} fps",
                        )
                    self.be_fast_multiplication = None

                self.be_fast = True

        # Monitor the frame fps
        self.queue.put(self.fps.value)

        # Now we know how many times differ from the ideal Frame Rate
        if differ <= 0:
            # raise ValueError('cannot maintain desired FPS fps')
            if self.be_fast:
                self.fps.value -= (
                    self.fps.fps_max_increment * self.be_fast_multiplication / 100
                )
            else:
                self.fps.value -= self.fps.fps_increment
            # Return False that because we haven't respected the ideal frame fps
            return False

        # Everything is fine , we have spare time then we can sleep for the rest of the frame time
        sleep(differ)

        if self.be_fast:
            self.fps.value += (
                self.fps.fps_max_increment * self.be_fast_multiplication / 100
            )
        else:
            self.fps.value += self.fps.fps_increment

        # Return True that because we have respect the ideal frame fps
        return True

    @property
    def fps_accelerated(self):
        return (self.fps.fps_max_increment * self.be_fast_multiplication) / 100

    @property
    def time(self):
        """
        Time should be take as a serious thing, you should try to impose only one time source in you program, then the
        :class:`Timer <glxeveloop.timer.Timer>` Class provide it own method for get the time by it self.

        :return: Unix time
        :rtype: int
        """
        return time()

    @property
    def time_departure(self):
        """
        Return the value set by a :func:`Timer._set_departure_time() <glxeveloop.timer.Timer._set_departure_time()>`

        :return: return :py:data:`departure_time` property.
        :rtype: Unix time
        """
        return self.__time_departure

    @time_departure.setter
    def time_departure(self, time_value=None):
        """
        Store a :func:`Timer.get_time() <glxeveloop.timer.Timer.get_time()>` return value inside
        :py:data:`departure_time` property.

        :param time_value: return value inside :py:data:`departure_time` property.
        :type time_value: float
        """
        if time_value and not isinstance(time_value, float):
            raise TypeError("'time_value' property value must be a float type or None")
        if self.time_departure != time_value:
            self.__time_departure = time_value

    @property
    def be_fast(self):
        """
        Return the value set by :func:`Timer._set_be_fast() <glxeveloop.timer.Timer._set_be_fast()>` method.

        You can set :py:attr:`__be_fast` property with

        :func:`Timer._set_be_fast() <glxeveloop.timer.Timer._set_be_fast()>` method.
        :return: :py:attr:`__be_fast` property
        :rtype: bool
        """
        return self.__be_fast

    @be_fast.setter
    def be_fast(self, be_fast):
        """
        Set the __be_fast property after have checked if teh value is different and if type is a boolean value like:
        True, False, O, 1

        That value will be used for fast convergence, when the Timer search for the best Frame fps

        :param be_fast:
        :type be_fast: bool
        :raise TypeError: if ``be_fast`` parameter is not a :py:data:`bool` type
        """
        if be_fast is None:
            be_fast = False
        if not isinstance(be_fast, bool):
            raise TypeError("'be_fast' parameter must be a bool")
        if self.be_fast is not be_fast:
            self.__be_fast = be_fast

    @property
    def be_fast_multiplication(self):
        """
        Return the be_fast_multiplication property value

        :return: :py:attr:`be_fast_multiplication` property
        :rtype: int
        """
        return self.__be_fast_multiplication

    @be_fast_multiplication.setter
    def be_fast_multiplication(self, value=None):
        """
        Set the :py:attr:`be_fast_multiplication` property

        That value will be used for fast convergence, when the Timer search for the best Frame fps

        :param value:
        :type value: int
        :raise TypeError: if ``be_fast_multiplication``  property value is is not a :py:data:`int` type or None
        """
        if value is None:
            value = 10
        if not isinstance(value, int):
            raise TypeError("'be_fast' parameter must be a int")
        if self.be_fast_multiplication != value:
            self.__be_fast_multiplication = value



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import codecs
import os

from setuptools import setup


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("APPLICATION_VERSION"):
            delimiter = '"' if '"' in line else "'"
            return line.split(delimiter)[1]
    else:
        raise RuntimeError("Unable to find version string.")


if __name__ == "__main__":
    if os.environ.get("APPLICATION_VERSION"):
        version = os.environ["APPLICATION_VERSION"]
    else:
        version = get_version("glxeveloop/__init__.py")
    print(f"Setup version is: {version}")
    setup(version=version)



import logging
from threading import Thread
from queue import Queue
from glxeveloop.hooks import Hooks


class HandleEvents:
    debug: bool
    hooks: Hooks
    queue: Queue

    def handle_event(self):
        event = None
        handler_list = []
        try:
            while not self.queue.empty():
                event = self.queue.get()
                if self.hooks.dispatch:
                    if hasattr(self, "debug") and self.debug:
                        logging.debug(
                            "%s",
                            f"{self.__class__.__name__}.handle_event ({event[0]}, {event[1]}) to {self.hooks.dispatch}",
                        )
                    handler_list.append(
                        Thread(target=self.hooks.dispatch(event[0], event[1]))
                    )

            for handler in handler_list:
                handler.start()
            for handler in handler_list:
                handler.join()

        except KeyError as the_error:  # pragma: no cover
            # Permit to have error logs about unknown event
            logging.error(
                "%s",
                f"{self.__class__.__name__}._handle_event(): KeyError:{the_error} event:{event}",
            )



import logging


class Stop:
    running: bool
    debug: bool

    def stop(self):
        """
        Stops a Loop from running. Any calls to run() for the loop will return.

        Note that sources that have already been dispatched when quit() is called will still be executed.

        .. :warning: A Loop quit() call will certainly cause the end of you programme
        """
        if hasattr(self, "debug") and self.debug:
            logging.debug("%s", f"{self.__class__.__name__}: Stopping")

        self.running = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from time import time, sleep
from queue import Queue
from glxeveloop.fps import FPS
from glxeveloop.timer import Timer


# Unittest
class TestTimer(unittest.TestCase):
    def setUp(self):
        self.timer = Timer()
        self.timer.debug = True

    def test_debug(self):
        self.timer.debug = False
        self.assertFalse(self.timer.debug)
        self.assertFalse(self.timer.fps.debug)

        self.timer.debug = True
        self.assertTrue(self.timer.debug)
        self.assertTrue(self.timer.fps.debug)

        self.assertRaises(TypeError, setattr, self.timer, "debug", "Hello.42")

    def test_default_values(self):
        self.assertEqual(self.timer.position, 0)
        # self.assertEqual(self.timer.queue.size, 10)
        self.assertEqual(self.timer.be_fast, False)
        self.assertEqual(self.timer.be_fast_multiplication, 10)

    def test_buffer(self):
        self.assertTrue(isinstance(self.timer.queue, Queue))
        old_buffer = self.timer.queue
        self.timer.queue = None
        self.assertTrue(isinstance(self.timer.queue, Queue))
        self.assertNotEqual(old_buffer, self.timer.queue)

        self.assertRaises(TypeError, setattr, self.timer, "queue", "Hello.42")

    def test_fps(self):
        self.assertTrue(isinstance(self.timer.fps, FPS))
        old_fps = self.timer.fps
        self.timer.fps = None
        self.assertTrue(isinstance(self.timer.fps, FPS))
        self.assertNotEqual(old_fps, self.timer.fps)

        self.timer.fps = FPS()
        self.assertTrue(isinstance(self.timer.fps, FPS))

        self.assertRaises(TypeError, setattr, self.timer, "fps", "Hello.42")

    def test_time(self):
        returned_value_1 = self.timer.time
        returned_value_2 = self.timer.time
        self.assertLessEqual(returned_value_1, returned_value_2)

    def test_time_departure(self):
        tested_value = time()
        self.timer.time_departure = tested_value
        self.assertEqual(self.timer.time_departure, tested_value)

        self.assertRaises(TypeError, setattr, self.timer, "time_departure", 42)

    def test_be_bast(self):
        value_tested = True
        self.timer.be_fast = value_tested
        value_returned = self.timer.be_fast
        self.assertEqual(value_tested, value_returned)

        self.assertRaises(TypeError, setattr, self.timer, "be_fast", "Hello.42")

    def test_be_fast_multiplication(self):
        self.timer.be_fast_multiplication = 42
        self.assertEqual(self.timer.be_fast_multiplication, 42)

        self.assertRaises(
            TypeError,
            setattr,
            self.timer,
            "be_fast_multiplication",
            str("Hello World!"),
        )

    def test_tick(self):
        self.timer.debug = True
        self.timer.position = 0
        self.timer.fps.value = 5
        self.timer.fps.max = 42

        count = 0
        while count < 42:
            self.timer.tick()
            sleep(0.1)
            count += 1

        count = 0
        while count < 42:
            self.timer.tick()
            count += 1

        count = 0
        self.timer.fps.value = 200
        while count < 42:
            self.timer.tick()
            sleep(0.01)
            count += 1

        count = 0
        while count < 42:
            self.timer.tick()
            sleep(0.01)
            count += 1

        self.timer.fps.value = 200
        while not self.timer.queue.empty():
            self.timer.queue.get()

        self.timer.queue.put(420.0)
        self.timer.queue.put(420.0)
        self.timer.queue.put(420.0)
        self.timer.queue.put(420.0)
        self.timer.queue.put(420.0)
        self.timer.queue.put(42.0)
        self.timer.queue.put(42.0)
        self.timer.queue.put(42.0)
        self.timer.queue.put(42.0)
        self.timer.queue.put(42.0)

        count = 0
        self.timer.fps.value = 40
        self.timer.be_fast = True
        while count < 42:
            self.timer.tick()
            sleep(0.01)
            count += 1


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import threading
import logging

from glxeveloop import MainLoop
from glxeveloop.properties.debug import DebugProperty


class Bus(DebugProperty):
    """
    :Description:

    The ``EventBusClient`` object is The bus it interconnects Widget
    """

    def __init__(self):
        DebugProperty.__init__(self)

        # Public attribute
        self.__subscriptions = {}
        # self.subscriptions = None

    @property
    def subscriptions(self) -> dict:
        """
        Return the subscriptions list

        :return: event buffer
        :rtype: dict
        """
        return self.__subscriptions

    @subscriptions.setter
    def subscriptions(self, value=None):
        if value is None:
            value = {}
        if not isinstance(value, dict):
            raise TypeError("'subscriptions' property value must be dict type or None")
        if self.subscriptions != value:
            self.__subscriptions = value

    def connect(self, detailed_signal: object, handler: object, *args: object):
        """
        The connect() method adds a function or method (handler) to the end of the event list
        for the named detailed_signal but before the default class signal handler.
        An optional set of parameters may be specified after the handler parameter.
        These will all be passed to the signal handler when invoked.

        :param detailed_signal: a string containing the signal name
        :type detailed_signal: str
        :param handler: a function handler
        :type handler: handler
        :param args: additional parameters arg1, arg2
        :type args: tuple
        """

        # If args is still None replace it by an empty list
        # if args is None:
        #     args = []

        # If detailed_signal is not in the event list create it
        if detailed_signal not in self.subscriptions:
            self.subscriptions[detailed_signal] = []

        self.subscriptions[detailed_signal].append(handler)

        if args:
            self.subscriptions[detailed_signal].append(args)

    def disconnect(self, detailed_signal, handler):
        """
        The disconnect() method removes the signal handler with the specified handler
        from the list of signal handlers for the object.

        :param detailed_signal: a string containing the signal name
        :type detailed_signal: str
        :param handler: a function handler
        :type handler: handler
        """
        if (
            detailed_signal in self.subscriptions
            and handler in self.subscriptions[detailed_signal]
        ):
            del self.subscriptions[detailed_signal]

    @staticmethod
    def emit(detailed_signal, args):
        """
        Emit signal in direction to the Mainloop.

        :param detailed_signal: a string containing the signal name
        :type detailed_signal: str
        :param args: additional parameters arg1, arg2
        :type args: dict
        """
        MainLoop().loop.queue.put([detailed_signal, args])

    def events_flush(self, detailed_signal, args):
        if detailed_signal in self.subscriptions:
            handler_list = []
            for handler in self.subscriptions[detailed_signal]:
                if self.debug:
                    logging.debug(
                        "%s",
                        f"{self.__class__.__name__}.events_dispatch({detailed_signal}, {args})",
                    )
                handler_list.append(
                    threading.Thread(target=handler(self, detailed_signal, args))
                )
            for handler in handler_list:
                handler.start()
            for handler in handler_list:
                handler.join()

    def events_dispatch(self, detailed_signal, args):
        """
        Flush Mainloop event to Child's father's for a Widget's recursive event dispatch

        :param detailed_signal: a string containing the signal name
        :type detailed_signal: str
        :param args: additional parameters arg1, arg2
        :type args: list
        """
        self.events_flush(detailed_signal, args)

        if hasattr(self, "eveloop_dispatch"):
            self.eveloop_dispatch(detailed_signal, args)



import unittest

from glxeveloop.loop.builder import Builder
from glxeveloop.loop.start import Start
from glxeveloop.loop.stop import Stop
from glxeveloop.loop.handle_events import HandleEvents
from glxeveloop.loop.sequence import SequenceLoop
from glxeveloop.hooks import Hooks
from queue import Queue
from glxeveloop.timer import Timer


class TestBuilders(unittest.TestCase):
    def test_Builder(self):
        builder = Builder().loop(
            debug=True,
            hooks=Hooks(),
            queue=Queue(),
            timer=Timer(),
            running=True,
            start=Start,
            stop=Stop,
            event_handler=HandleEvents,
            sequence=SequenceLoop,
        )


if __name__ == "__main__":
    unittest.main()



from glxeveloop.properties.debug import DebugProperty
from glxeveloop.properties.hooks import HooksProperty
from glxeveloop.properties.queue import QueueProperty
from glxeveloop.properties.timer import TimerProperty
from glxeveloop.properties.running import RunningProperty
from glxeveloop.loop.start import Start
from glxeveloop.loop.stop import Stop
from glxeveloop.loop.handle_events import HandleEvents
from glxeveloop.loop.sequence import SequenceLoop


class Loop(
    DebugProperty,
    QueueProperty,
    HooksProperty,
    RunningProperty,
    TimerProperty,
    Start,
    Stop,
    HandleEvents,
    SequenceLoop,
):
    def __init__(self):
        DebugProperty.__init__(self)
        QueueProperty.__init__(self)
        HooksProperty.__init__(self)
        RunningProperty.__init__(self)
        TimerProperty.__init__(self)
        Start.__init__(self)
        Stop.__init__(self)
        HandleEvents.__init__(self)
        SequenceLoop.__init__(self)



from glxeveloop.loop.loop import Loop


class Builder:
    @staticmethod
    def loop(
        debug=None,
        hooks=None,
        queue=None,
        timer=None,
        running=None,
        start=None,
        stop=None,
        event_handler=None,
        sequence=None,
    ) -> object:
        new_loop = Loop()

        if debug and hasattr(new_loop, "debug"):
            new_loop.debug = debug
        if hooks and hasattr(new_loop, "hooks"):
            new_loop.hooks = hooks
        if queue and hasattr(new_loop, "queue"):
            new_loop.queue = queue
        if queue and hasattr(new_loop, "timer"):
            new_loop.timer = timer
        if running and hasattr(new_loop, "running"):
            new_loop.running = running
        if start and hasattr(new_loop, "start"):
            new_loop.start = start
        if stop and hasattr(new_loop, "stop"):
            new_loop.stop = stop
        if event_handler and hasattr(new_loop, "handle_event"):
            new_loop.handle_event = event_handler
        if sequence and hasattr(new_loop, "sequence"):
            new_loop.sequence = sequence

        return new_loop



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# It script it publish under GNU GENERAL PUBLIC LICENSE
# http://www.gnu.org/licenses/gpl-3.0.en.html
# Author: Tuuux <tuxa at rtnp dot org> all rights reserved

import logging
from glxeveloop.properties.debug import DebugProperty


class FPS(DebugProperty):
    def __init__(self):
        DebugProperty.__init__(self)
        self.__value = None
        self.__fps_increment = None
        self.__min = None
        self.__fps_min_increment = None
        self.__max = None
        self.__fps_max_increment = None

        self.min = None
        self.max = None
        self.value = None

        self.fps_min_increment = None
        self.fps_max_increment = None
        self.fps_increment = None

    @property
    def value(self):
        """
        The property :py:data:`value`, store the actual speed rate of the MainLoop, it property is use by
        the class :class:`Timer <glxeveloop.timer.Timer>`, and will be updated at each loop iteration.

        By set :py:data:`value` property before call :func:`MainLoop.start() <glxeveloop.MainLoop.start>` , it imposes to start with it value, and slide to the
        :py:data:`max`

        .. note:: If set the value will be clamped between :py:data:`min` and :py:data:`max` properties value.

        .. warning:: That not correspond exactly to a true movies or game FPS, it corresponds more about how many \
        times **1 second** is divided..

        :return: :py:data:`value` property. (in **rate**)
        :rtype: :py:data:`float` or :py:data:`int`
        :raise TypeError: if parameter is not a :py:data:`float`|:py:data:`int` type or None
        """
        return self.__value

    @value.setter
    def value(self, value=None):
        """
        Set the :py:data:`value` property.

        :param value: Frames number per second. (in **fps**)
        :type value: float, int or None
        """
        if value is None:
            value = 60.00
        if isinstance(value, int):
            value = float(value)
        if not isinstance(value, float):
            raise TypeError("'value' property value must be float/int type or None")

        # CLAMP to the absolute value
        clamped_value = abs(max(min(self.max, value), self.min))

        if self.value != round(clamped_value, 2):
            self.__value = round(clamped_value, 2)
            if self.debug:
                logging.debug("{%s}", f"{self.__class__.__name__}.value({self.value})")

    @property
    def min(self):
        """
        The min Frames number per second allowed before the :class:`Timer <glxeveloop.timer.Timer>` stop to apply
         a rate limit to the :py:data:`value` property. (in **rate**)

        It can be considered as the min value of the CLAMP process


        :return: :py:data:`min` property value. (in **fps**)
        :rtype: float
        """
        return self.__min

    @min.setter
    def min(self, value=None):
        """
        Set the :py:data:`min` property value.

        It corresponds to an imposed minimal amount of frame rate

        :param value: :py:data:`min` property value. (in **fps**)
        :type value: float
        :raise TypeError: if ``min`` parameter is not a :py:data:`float` type
        """
        if value is None:
            value = 30.0

        if isinstance(value, int):
            value = float(value)

        if not isinstance(value, float):
            raise TypeError("'min' property value must be float/int type or None")

        if self.min != value:
            self.__min = value
            if self.debug:
                logging.debug("{%s}", f"{self.__class__.__name__}.min({self.min})")

    @property
    def max(self):
        """
        The maximum Frames number per second allowed before the :class:`Timer <glxeveloop.timer.Timer>` start to rate
         limit the :py:data:`rate` property.

        It can be considered as the max value of the CLAMP process.

        .. note:: By default, it has no limit fps_max = float("inf")



        :return: :py:attr:`fps_max` property value. (in **fps**)
        :rtype: float
        """
        return self.__max

    @max.setter
    def max(self, value=None):
        """
        Set the :class:`Timer <glxeveloop.timer.Timer>` :py:attr:`fps_max` property value.

        It corresponds to an imposed max amount of frame rate used during acceleration phase

        :param value: :py:attr:`fps_max` property value. (in **rate**)
        :type value: :py:obj:`float` or :py:data:`int` or :py:obj:`None`
        :raise TypeError: if ``max_fps`` parameter value is not a :py:data:`float` or
        :py:data:`int` type or :py:obj:`None`
        """
        if value is None:
            value = float("inf")

        if isinstance(value, int):
            value = float(value)

        if not isinstance(value, float):
            raise TypeError(
                "'max_fps' property value parameter must be a float or None"
            )

        if self.max != value:
            self.__max = value
            if self.debug:
                logging.debug("{%s}", f"{self.__class__.__name__}.max({self.value})")

    @property
    def fps_increment(self):
        """
        Get the :class:`Timer <glxeveloop.timer.Timer>` :py:data:`fps_increment` property value.

        :return: :py:data:`fps_increment` property value. (in **fps**)
        :rtype: float
        """
        return self.__fps_increment

    @fps_increment.setter
    def fps_increment(self, fps_increment=0.1):
        """
        Set the :class:`Timer <glxeveloop.timer.Timer>` :py:attr:`fps_increment` property.

        The self-correcting timing algorithms will try to increase or decrease :py:attr:`fps_increment` property
        with it step increment.

        :param fps_increment: Frames number per second. (in **rate**)
        :type fps_increment: float
        :raise TypeError: if ``fps_increment`` parameter is not a :py:data:`float` type
        """
        if fps_increment is None:
            fps_increment = 1.0
        if not isinstance(fps_increment, float):
            raise TypeError("'rate' parameter must be a float")
        if self.fps_increment != fps_increment:
            self.__fps_increment = fps_increment
            if self.debug:
                logging.debug(
                    "{%s}",
                    f"{self.__class__.__name__}.fps_increment({self.fps_increment})",
                )

    @property
    def fps_min_increment(self):
        """
        Get the smaller of step increment

        The :class:`Timer <glxeveloop.timer.Timer>` :py:data:`fps_min_increment` property value.

        See :func:`Timer.set_fps_min_increment() <glxeveloop.timer.Timer.set_fps_min_increment()>`
        for more information's

        :return: :py:data:`fps_min_increment` property value. (in **rate**)
        :rtype: float
        """
        return self.__fps_min_increment

    @fps_min_increment.setter
    def fps_min_increment(self, fps_min_increment=0.1):
        """
        Set the :class:`Timer <glxeveloop.timer.Timer>` :py:data:`fps_min_increment` increment.

        The algorithms will try to increase or decrease :py:data:`rate` property with
        :py:attr:`fps_increment` as step .

        For fast limit rate stabilization the :class:`Timer <glxeveloop.timer.Timer>` can use
        :py:data:`fps_min_increment`
        and :py:data:`fps_max_increment` property for make a gap in contain in a range, where
        :py:data:`fps_min_increment` will force a minimal amount of increment and
        :py:data:`fps_max_increment` will force a maximal amount of increment.

        :param fps_min_increment: Frames number per second. (in **rate**)
        :type fps_min_increment: float
        :raise TypeError: if ``fps_min_increment`` parameter is not a :py:data:`float` type
        """
        if fps_min_increment is None:
            fps_min_increment = 0.1
        if not isinstance(fps_min_increment, float):
            raise TypeError("'fps_min_increment' parameter must be a float type")
        if self.fps_min_increment != fps_min_increment:
            self.__fps_min_increment = fps_min_increment
            if self.debug:
                logging.debug(
                    "{%s}",
                    f"{self.__class__.__name__}.fps_min_increment({self.fps_min_increment})",
                )

    @property
    def fps_max_increment(self):
        """
        :py:data:`fps_max_increment` is the upper allowed increment value

        The self-correcting timing will try to adjust :py:data:`value` property in range of
        :py:data:`fps_min_increment` to :py:data:`fps_max_increment`

        **Default Value::** 100.0

        :return: :class:`Timer <glxeveloop.timer.Timer>` :py:data:`fps_max_increment` property value. (in **rate**)
        :rtype: float
        :raise TypeError: if ``fps_max_increment`` parameter is not a :py:data:`float` type
        """
        return self.__fps_max_increment

    @fps_max_increment.setter
    def fps_max_increment(self, fps_max_increment=None):
        if fps_max_increment is None:
            fps_max_increment = 100.0
        if not isinstance(fps_max_increment, float):
            raise TypeError("'max_fps_increment' parameter must be a float")
        if self.fps_max_increment != fps_max_increment:
            self.__fps_max_increment = fps_max_increment
            if self.debug:
                logging.debug(
                    "{%s}",
                    f"{self.__class__.__name__}.fps_max_increment({self.fps_max_increment})",
                )
